@@ -218,16 +218,13 @@ struct lttng_ust_tracepoint_dlopen {
 	int (*tracepoint_register_lib)(struct lttng_ust_tracepoint * const *tracepoints_start,
 		int tracepoints_count);
 	int (*tracepoint_unregister_lib)(struct lttng_ust_tracepoint * const *tracepoints_start);
-#ifndef _LGPL_SOURCE
 	void (*rcu_read_lock_sym_bp)(void);
 	void (*rcu_read_unlock_sym_bp)(void);
 	void *(*rcu_dereference_sym_bp)(void *p);
-#endif
 };
 
 extern struct lttng_ust_tracepoint_dlopen tracepoint_dlopen;
-
-#if defined(TRACEPOINT_DEFINE) || defined(TRACEPOINT_CREATE_PROBES)
+extern struct lttng_ust_tracepoint_dlopen *tracepoint_dlopen_ptr;
 
 /*
  * These weak symbols, the constructor, and destructor take care of
@@ -240,6 +237,49 @@ int __tracepoint_ptrs_registered
 	__attribute__((weak, visibility("hidden")));
 struct lttng_ust_tracepoint_dlopen tracepoint_dlopen
 	__attribute__((weak, visibility("hidden")));
+/*
+ * Deal with gcc O1 optimisation issues with weak hidden symbols. gcc
+ * 4.8 and prior does not have the same behavior for symbol scoping on
+ * 32-bit powerpc depending on the object size: symbols for objects of 8
+ * bytes or less have the same address throughout a module, whereas they
+ * have different addresses between compile units for objects larger
+ * than 8 bytes. Add this pointer indirection to ensure that the symbol
+ * scoping match that of the other weak hidden symbols found in this
+ * header.
+ */
+struct lttng_ust_tracepoint_dlopen *tracepoint_dlopen_ptr
+	__attribute__((weak, visibility("hidden")));
+
+/*
+ * Tracepoint dynamic linkage handling (callbacks). Hidden visibility: shared
+ * across objects in a module/main executable. The callbacks are used to
+ * control and check if the destructors should be executed.
+ */
+struct lttng_ust_tracepoint_destructors_syms {
+	int *old_tracepoint_disable_destructors;
+	void (*tracepoint_disable_destructors)(void);
+	int (*tracepoint_get_destructors_state)(void);
+};
+
+extern struct lttng_ust_tracepoint_destructors_syms tracepoint_destructors_syms;
+extern struct lttng_ust_tracepoint_destructors_syms *tracepoint_destructors_syms_ptr;
+
+struct lttng_ust_tracepoint_destructors_syms tracepoint_destructors_syms
+	__attribute__((weak, visibility("hidden")));
+struct lttng_ust_tracepoint_destructors_syms *tracepoint_destructors_syms_ptr
+	__attribute__((weak, visibility("hidden")));
+
+static inline void tracepoint_disable_destructors(void)
+{
+	if (!tracepoint_dlopen_ptr)
+		tracepoint_dlopen_ptr = &tracepoint_dlopen;
+	if (!tracepoint_destructors_syms_ptr)
+		tracepoint_destructors_syms_ptr = &tracepoint_destructors_syms;
+	if (tracepoint_dlopen_ptr->liblttngust_handle
+			&& tracepoint_destructors_syms_ptr->tracepoint_disable_destructors)
+		tracepoint_destructors_syms_ptr->tracepoint_disable_destructors();
+	*tracepoint_destructors_syms_ptr->old_tracepoint_disable_destructors = 1;
+}
 
 #ifndef _LGPL_SOURCE
 static inline void lttng_ust_notrace
@@ -247,24 +287,26 @@ __tracepoint__init_urcu_sym(void);
 static inline void
 __tracepoint__init_urcu_sym(void)
 {
+	if (!tracepoint_dlopen_ptr)
+		tracepoint_dlopen_ptr = &tracepoint_dlopen;
 	/*
 	 * Symbols below are needed by tracepoint call sites and probe
 	 * providers.
 	 */
-	if (!tracepoint_dlopen.rcu_read_lock_sym_bp)
-		tracepoint_dlopen.rcu_read_lock_sym_bp =
+	if (!tracepoint_dlopen_ptr->rcu_read_lock_sym_bp)
+		tracepoint_dlopen_ptr->rcu_read_lock_sym_bp =
 			URCU_FORCE_CAST(void (*)(void),
-				dlsym(tracepoint_dlopen.liblttngust_handle,
+				dlsym(tracepoint_dlopen_ptr->liblttngust_handle,
 					"tp_rcu_read_lock_bp"));
-	if (!tracepoint_dlopen.rcu_read_unlock_sym_bp)
-		tracepoint_dlopen.rcu_read_unlock_sym_bp =
+	if (!tracepoint_dlopen_ptr->rcu_read_unlock_sym_bp)
+		tracepoint_dlopen_ptr->rcu_read_unlock_sym_bp =
 			URCU_FORCE_CAST(void (*)(void),
-				dlsym(tracepoint_dlopen.liblttngust_handle,
+				dlsym(tracepoint_dlopen_ptr->liblttngust_handle,
 					"tp_rcu_read_unlock_bp"));
-	if (!tracepoint_dlopen.rcu_dereference_sym_bp)
-		tracepoint_dlopen.rcu_dereference_sym_bp =
+	if (!tracepoint_dlopen_ptr->rcu_dereference_sym_bp)
+		tracepoint_dlopen_ptr->rcu_dereference_sym_bp =
 			URCU_FORCE_CAST(void *(*)(void *p),
-				dlsym(tracepoint_dlopen.liblttngust_handle,
+				dlsym(tracepoint_dlopen_ptr->liblttngust_handle,
 					"tp_rcu_dereference_sym_bp"));
 }
 #else
@@ -281,13 +323,19 @@ __tracepoints__init(void);
 static void
 __tracepoints__init(void)
 {
-	if (__tracepoint_registered++)
+	if (__tracepoint_registered++) {
+		if (!tracepoint_dlopen_ptr->liblttngust_handle)
+			return;
+		__tracepoint__init_urcu_sym();
 		return;
+	}
 
-	if (!tracepoint_dlopen.liblttngust_handle)
-		tracepoint_dlopen.liblttngust_handle =
+	if (!tracepoint_dlopen_ptr)
+		tracepoint_dlopen_ptr = &tracepoint_dlopen;
+	if (!tracepoint_dlopen_ptr->liblttngust_handle)
+		tracepoint_dlopen_ptr->liblttngust_handle =
 			dlopen("liblttng-ust-tracepoint.so.0", RTLD_NOW | RTLD_GLOBAL);
-	if (!tracepoint_dlopen.liblttngust_handle)
+	if (!tracepoint_dlopen_ptr->liblttngust_handle)
 		return;
 	__tracepoint__init_urcu_sym();
 }
@@ -301,17 +349,46 @@ __tracepoints__destroy(void)
 
 	if (--__tracepoint_registered)
 		return;
-	if (tracepoint_dlopen.liblttngust_handle && !__tracepoint_ptrs_registered) {
-		ret = dlclose(tracepoint_dlopen.liblttngust_handle);
-		if (ret) {
-			fprintf(stderr, "Error (%d) in dlclose\n", ret);
-			abort();
-		}
-		memset(&tracepoint_dlopen, 0, sizeof(tracepoint_dlopen));
+	if (!tracepoint_dlopen_ptr)
+		tracepoint_dlopen_ptr = &tracepoint_dlopen;
+	if (!tracepoint_destructors_syms_ptr)
+		tracepoint_destructors_syms_ptr = &tracepoint_destructors_syms;
+	if (!tracepoint_dlopen_ptr->liblttngust_handle)
+		return;
+	if (__tracepoint_ptrs_registered)
+		return;
+	/*
+	 * Lookup if destructors must be executed using the new method.
+	 */
+	if (tracepoint_destructors_syms_ptr->tracepoint_get_destructors_state
+		&& !tracepoint_destructors_syms_ptr->tracepoint_get_destructors_state()) {
+		/*
+		 * The tracepoint_get_destructors_state symbol was found with
+		 * dlsym but its returned value is 0 meaning that destructors
+		 * must not be executed.
+		 */
+		return;
+	}
+	/*
+	 * Lookup if destructors must be executed using the old method.
+	 */
+	if (tracepoint_destructors_syms_ptr->old_tracepoint_disable_destructors
+		&& *tracepoint_destructors_syms_ptr->old_tracepoint_disable_destructors) {
+		/*
+		 * The old_tracepoint_disable_destructors symbol was found with
+		 * dlsym but its value is 1 meaning that destructors must not
+		 * be executed.
+		 */
+		return;
 	}
-}
 
-#endif
+	ret = dlclose(tracepoint_dlopen_ptr->liblttngust_handle);
+	if (ret) {
+		fprintf(stderr, "Error (%d) in dlclose\n", ret);
+		abort();
+	}
+	memset(tracepoint_dlopen_ptr, 0, sizeof(*tracepoint_dlopen_ptr));
+}
 
 #ifdef TRACEPOINT_DEFINE
 
@@ -374,22 +451,38 @@ __tracepoints__ptrs_init(void)
 {
 	if (__tracepoint_ptrs_registered++)
 		return;
-	if (!tracepoint_dlopen.liblttngust_handle)
-		tracepoint_dlopen.liblttngust_handle =
+	if (!tracepoint_dlopen_ptr)
+		tracepoint_dlopen_ptr = &tracepoint_dlopen;
+	if (!tracepoint_dlopen_ptr->liblttngust_handle)
+		tracepoint_dlopen_ptr->liblttngust_handle =
 			dlopen("liblttng-ust-tracepoint.so.0", RTLD_NOW | RTLD_GLOBAL);
-	if (!tracepoint_dlopen.liblttngust_handle)
+	if (!tracepoint_dlopen_ptr->liblttngust_handle)
 		return;
-	tracepoint_dlopen.tracepoint_register_lib =
+	if (!tracepoint_destructors_syms_ptr)
+		tracepoint_destructors_syms_ptr = &tracepoint_destructors_syms;
+	tracepoint_dlopen_ptr->tracepoint_register_lib =
 		URCU_FORCE_CAST(int (*)(struct lttng_ust_tracepoint * const *, int),
-				dlsym(tracepoint_dlopen.liblttngust_handle,
+				dlsym(tracepoint_dlopen_ptr->liblttngust_handle,
 					"tracepoint_register_lib"));
-	tracepoint_dlopen.tracepoint_unregister_lib =
+	tracepoint_dlopen_ptr->tracepoint_unregister_lib =
 		URCU_FORCE_CAST(int (*)(struct lttng_ust_tracepoint * const *),
-				dlsym(tracepoint_dlopen.liblttngust_handle,
+				dlsym(tracepoint_dlopen_ptr->liblttngust_handle,
 					"tracepoint_unregister_lib"));
+	tracepoint_destructors_syms_ptr->old_tracepoint_disable_destructors =
+		URCU_FORCE_CAST(int *,
+				dlsym(tracepoint_dlopen_ptr->liblttngust_handle,
+					"__tracepoints__disable_destructors"));
+	tracepoint_destructors_syms_ptr->tracepoint_disable_destructors =
+		URCU_FORCE_CAST(void (*)(void),
+				dlsym(tracepoint_dlopen_ptr->liblttngust_handle,
+					"tp_disable_destructors"));
+	tracepoint_destructors_syms_ptr->tracepoint_get_destructors_state =
+		URCU_FORCE_CAST(int (*)(void),
+				dlsym(tracepoint_dlopen_ptr->liblttngust_handle,
+					"tp_get_destructors_state"));
 	__tracepoint__init_urcu_sym();
-	if (tracepoint_dlopen.tracepoint_register_lib) {
-		tracepoint_dlopen.tracepoint_register_lib(__start___tracepoints_ptrs,
+	if (tracepoint_dlopen_ptr->tracepoint_register_lib) {
+		tracepoint_dlopen_ptr->tracepoint_register_lib(__start___tracepoints_ptrs,
 				__stop___tracepoints_ptrs -
 				__start___tracepoints_ptrs);
 	}
@@ -404,15 +497,22 @@ __tracepoints__ptrs_destroy(void)
 
 	if (--__tracepoint_ptrs_registered)
 		return;
-	if (tracepoint_dlopen.tracepoint_unregister_lib)
-		tracepoint_dlopen.tracepoint_unregister_lib(__start___tracepoints_ptrs);
-	if (tracepoint_dlopen.liblttngust_handle && !__tracepoint_registered) {
-		ret = dlclose(tracepoint_dlopen.liblttngust_handle);
+	if (!tracepoint_dlopen_ptr)
+		tracepoint_dlopen_ptr = &tracepoint_dlopen;
+	if (!tracepoint_destructors_syms_ptr)
+		tracepoint_destructors_syms_ptr = &tracepoint_destructors_syms;
+	if (tracepoint_dlopen_ptr->tracepoint_unregister_lib)
+		tracepoint_dlopen_ptr->tracepoint_unregister_lib(__start___tracepoints_ptrs);
+	if (tracepoint_dlopen_ptr->liblttngust_handle
+			&& tracepoint_destructors_syms_ptr->tracepoint_get_destructors_state
+			&& tracepoint_destructors_syms_ptr->tracepoint_get_destructors_state()
+			&& !__tracepoint_ptrs_registered) {
+		ret = dlclose(tracepoint_dlopen_ptr->liblttngust_handle);
 		if (ret) {
 			fprintf(stderr, "Error (%d) in dlclose\n", ret);
 			abort();
 		}
-		memset(&tracepoint_dlopen, 0, sizeof(tracepoint_dlopen));
+		memset(tracepoint_dlopen_ptr, 0, sizeof(*tracepoint_dlopen_ptr));
 	}
 }
 
@@ -430,6 +530,50 @@ __tracepoints__ptrs_destroy(void)
 
 /* The following declarations must be outside re-inclusion protection. */
 
+#ifndef TRACEPOINT_ENUM
+
+/*
+ * Tracepoint Enumerations
+ *
+ * The enumeration is a mapping between an integer, or range of integers, and
+ * a string. It can be used to have a more compact trace in cases where the
+ * possible values for a field are limited:
+ *
+ * An example:
+ *
+ * TRACEPOINT_ENUM(someproject_component, enumname,
+ *	TP_ENUM_VALUES(
+ *		ctf_enum_value("even", 0)
+ *		ctf_enum_value("uneven", 1)
+ *		ctf_enum_range("twoto4", 2, 4)
+ *		ctf_enum_value("five", 5)
+ *	)
+ * )
+ *
+ * Where "someproject_component" is the name of the component this enumeration
+ * belongs to and "enumname" identifies this enumeration. Inside the
+ * TP_ENUM_VALUES macro is the actual mapping. Each string value can map
+ * to either a single value with ctf_enum_value or a range of values
+ * with ctf_enum_range.
+ *
+ * Enumeration ranges may overlap, but the behavior is implementation-defined,
+ * each trace reader will handle overlapping as it wishes.
+ *
+ * That enumeration can then be used in a field inside the TP_FIELD macro using
+ * the following line:
+ *
+ * ctf_enum(someproject_component, enumname, enumtype, enumfield, enumval)
+ *
+ * Where "someproject_component" and "enumname" match those in the
+ * TRACEPOINT_ENUM, "enumtype" is a signed or unsigned integer type
+ * backing the enumeration, "enumfield" is the name of the field and
+ * "enumval" is the value.
+ */
+
+#define TRACEPOINT_ENUM(provider, name, values)
+
+#endif /* #ifndef TRACEPOINT_ENUM */
+
 #ifndef TRACEPOINT_EVENT
 
 /*
@@ -453,6 +597,9 @@ __tracepoints__ptrs_destroy(void)
  *         * Integer, printed with 0x base 16 * 
  *         ctf_integer_hex(unsigned long, field_d, arg1)
  *
+ *         * Enumeration *
+ *         ctf_enum(someproject_component, enum_name, int, field_e, arg0)
+ *
  *         * Array Sequence, printed as UTF8-encoded array of bytes * 
  *         ctf_array_text(char, field_b, string, FIXED_LEN)
  *         ctf_sequence_text(char, field_c, string, size_t, strlen)
