@@ -55,7 +55,8 @@ static inline
 size_t record_header_size(const struct lttng_ust_lib_ring_buffer_config *config,
 				 struct channel *chan, size_t offset,
 				 size_t *pre_header_padding,
-				 struct lttng_ust_lib_ring_buffer_ctx *ctx)
+				 struct lttng_ust_lib_ring_buffer_ctx *ctx,
+				 void *client_ctx)
 {
 	return 0;
 }
@@ -72,7 +73,8 @@ static
 size_t client_record_header_size(const struct lttng_ust_lib_ring_buffer_config *config,
 				 struct channel *chan, size_t offset,
 				 size_t *pre_header_padding,
-				 struct lttng_ust_lib_ring_buffer_ctx *ctx)
+				 struct lttng_ust_lib_ring_buffer_ctx *ctx,
+				 void *client_ctx)
 {
 	return 0;
 }
@@ -206,7 +208,8 @@ struct lttng_channel *_channel_create(const char *name,
 				unsigned int read_timer_interval,
 				unsigned char *uuid,
 				uint32_t chan_id,
-				const int *stream_fds, int nr_stream_fds)
+				const int *stream_fds, int nr_stream_fds,
+				int64_t blocking_timeout)
 {
 	struct lttng_channel chan_priv_init;
 	struct lttng_ust_shm_handle *handle;
@@ -222,7 +225,7 @@ struct lttng_channel *_channel_create(const char *name,
 			&chan_priv_init,
 			buf_addr, subbuf_size, num_subbuf,
 			switch_timer_interval, read_timer_interval,
-			stream_fds, nr_stream_fds);
+			stream_fds, nr_stream_fds, blocking_timeout);
 	if (!handle)
 		return NULL;
 	lttng_chan = priv;
@@ -242,7 +245,7 @@ int lttng_event_reserve(struct lttng_ust_lib_ring_buffer_ctx *ctx, uint32_t even
 {
 	int ret;
 
-	ret = lib_ring_buffer_reserve(&client_config, ctx);
+	ret = lib_ring_buffer_reserve(&client_config, ctx, NULL);
 	if (ret)
 		return ret;
 	if (caa_likely(ctx->ctx_len
